//////////////////////////////////////////////////
//////////////////////////////////////////////////
// This source code is Copyright (2011) of      //
//  Jeppe R. Andersen and Jennifer M. Smillie   //
// and is distributed under the                 //
// Gnu Public License version 2                 //
// http://www.gnu.org/licenses/gpl-2.0.html     //
// You are allowed to distribute and alter the  //
// source under the conditions of the GPLv2     //
// as long as this copyright notice             //
// is unaltered and distributed with the source //
// Any use should comply with the               //
//             MCNET GUIDELINES                 //
//    for Event Generator Authors and Users     //
// as distributed with this source code         //
//////////////////////////////////////////////////
//////////////////////////////////////////////////
#include "JetGen.h"

extern "C" {
  void arjini_(int *, int *, double *, int *, double *);

  void arjshw_(double *, int *, double *, int *, int *, int *, const int *,
	       int *, double *, int *, int *, int *);

  void arjwgt_(double *, double *, int *, double *, double **,
	       int *, double *, double **, double *, double *);

  void setmsta_(int *, int *);

  void setstable_(int *, int *);

  void setpara_(int *, double *);

}

CMultijet * generator; // need this pointer for the subtraction
static const int maxp = 4000;

SGeneratorFlags Flags;
double jetrapmax(4.5);
double ptbarmin(0.);  //  Don't apply cut unless you set it by hand
bool SubtractFirstOnly(true); // Subtract only for emissions from original dipoles

/**
 * Routine called by Ariadne after each emission involving a gluon
 * being emitted between the two leading jets in rapidity. \a pg is
 * the momentum of the emitted gluon. If \a em1 or \a em2 are
 * non-negative they correspond to an emitter particle in the original
 * state (in the vectors given to arjshw_). If -1, they correspond to
 * a remnant particle and if -2 they correspond to a previously
 * emitted parton. \a pem1 and \a pem are the momenta of the emittor
 * partons after the emission. \a Psplit is the value of the splitting
 * function used by Ariadne, and should on return be set to the value
 * which should be subtracted from this splitting function.
 */
void arjwgt_(double * pg, double * hg, int * em1, double * pem1, double ** hem1,
	     int * em2, double * pem2, double ** hem2,
	     double * del, double * Psplit) {

  // std::cout << "ori " << pg[0] << " " << pg[1] << " " << pg[2] << " " << pg[3] << endl;
  // cout << "--- " << hg[0] << " " << hg[1]
  //      << " " << hg[2] << " " << hg[3] << endl;
  // cout << "--+ " << hg[5] << " " << hg[6]
  //      << " " << hg[7] << " " << hg[8] << endl;
  // cout << "-+- " << hg[10] << " " << hg[11]
  //      << " " << hg[12] << " " << hg[13] << endl;
  // cout << "-++ " << hg[15] << " " << hg[16]
  //      << " " << hg[17] << " " << hg[18] << endl;
  // cout << "+-- " << hg[20] << " " << hg[21]
  //      << " " << hg[22] << " " << hg[23] << endl;
  // cout << "+-+ " << hg[25] << " " << hg[26]
  //      << " " << hg[27] << " " << hg[28] << endl;
  // cout << "++- " << hg[30] << " " << hg[31]
  //      << " " << hg[32] << " " << hg[33] << endl;
  // cout << "+++ " << hg[35] << " " << hg[36]
  //      << " " << hg[37] << " " << hg[38] << endl;
  // cout << "123" << endl
  //      << "1 delta " << del[0] << " in log(pt^2)" << endl
  //      << "2 delta " << del[1] << " in y" << endl
  //      << "3 delta " << del[2] << " in phi" << endl;
  // cout << "Org. Psplit "<<*Psplit<<std::endl;
  std::vector<CLHEP::HepLorentzVector> testemissions;
  CLHEP::HepLorentzVector temp;
  
  // the original emission
  temp.set(pg[0],pg[1],pg[2],pg[3]);
  testemissions.push_back(temp);
  
#if 0 // 1: no subtraction
  *Psplit=0.;
#else  
  double subtract(0.);
  if (SubtractFirstOnly) {
    if (*em1>0&&*em2>0) 
      subtract=generator->SplitSubtract(testemissions,del);
  }
  else
    subtract=generator->SplitSubtract(testemissions,del);
  
  *Psplit=subtract;
#endif
}

int init(int argn, char* argv[])
{
  ifstream input;
  istringstream sline;
  int iline(0);
  string label,temp,line;

  if (argn!=2) {
    cerr << "# Usage:\n./JetGen [input.file]\n";
    cerr << "# where input.file is a file with parameters controlling the generation of events.\n";
    if (argn>2)
      return 1;
    else {
      cout << "# Using standard values for generator parameters\n";
      return 0;
    }
  } else {
    // Now open input file:
    input.open(argv[1], ios::in); // Open input file
    
    // Read parameters in from input file:
    while(getline(input,line)) {
      line.push_back(' ');
      sline.str(line);
      sline>>label;
      iline++;
      if(label=="Ebeam:") {
	if(!(sline>>Flags.Ebeam)) {
	  cout << "Error in input file at line " <<iline<<endl;
	  cout << "Value of Beam Energy invalid!"<<endl;
	  return 1;
	}
	if(Flags.Ebeam<0) {
	  cout << "Beam Energy<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="jetptmin:") {
	if(!(sline>>Flags.jetptmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of jetptmin invalid!"<<endl;
	  return 1;
	}
	if(Flags.jetptmin<0) {
	  cout <<"jetptmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="genjetptmin:") {
	if(!(sline>>Flags.genjetptmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of genjetptmin invalid!"<<endl;
	  return 1;
	}
	if(Flags.genjetptmin<0) {
	  cout <<"genjetptmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="extpartonptmin:") {
	if(!(sline>>Flags.extpartonptmin)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of extpartonptmin invalid!"<<endl;
	  return 1;
	}
	if(Flags.extpartonptmin<0) {
	  cout <<"extpartonptmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="scale:") {
	if(!(sline>>Flags.scale)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of scale invalid!"<<endl;
	  return 1;
	}
	if(Flags.scale<0) {
	  cout<<"scale<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="scalesetting:") { //0: fixed 1:PTMAX
	if(!(sline>>Flags.scalesetting)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of scalesetting invalid!"<<endl;
	  return 1;
	}
	if(Flags.scalesetting!=0&&Flags.scalesetting!=1) {
	  cout<<"scalesetting \notin {0,1}  at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="scalefactor:") { 
	if(!(sline>>Flags.scalefactor)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of scalefactor invalid!"<<endl;
	  return 1;
	}
	if(Flags.scalefactor<0) {
	  cout<<"scalefactor <0  at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="logcorrect:") { 
	if(!(sline>>Flags.logcorrect)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of logcorrect invalid!"<<endl;
	  return 1;
	}
	if(Flags.logcorrect!=0&&Flags.logcorrect!=1) {
	  cout<<"logcorrrect \notin {0,1}  at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="jetrapmax:") {
	if(!(sline>>jetrapmax)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of jetrapmax invalid!"<<endl;
	  return 1;
	}
	if(jetrapmax<0) {
	  cout<<"jetrapmax<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if (label=="ptbarmin:") {
	if (!(sline>>ptbarmin)) {
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of ptbarmin invalid!"<<endl;
	  return 1;
	}
	if (ptbarmin<0) {
	  cout<<"ptbarmin<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="NEvents:") {
	if(!(sline>>Flags.NEvents)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of NEvents invalid!"<<endl;
	  return 1;
	}
	if(Flags.NEvents<0) {
	  cout<<"NEvents<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="beam1:") {
	if(!(sline>>temp)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Beam 1 type invalid!"<<endl;
	  return 1;
	}
	if(temp=="P") Flags.beam[0]=1;
	else if(temp=="PBAR") Flags.beam[0]=-1;
	else {
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "Unrecognised beam type for beam 1!"<<endl;
	  return 1;
	}
      }
      else if(label=="beam2:") {
	if(!(sline>>temp)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Beam 2 type invalid!"<<endl;
	  return 1;
	}
	if(temp=="P") Flags.beam[1]=1;
	else if(temp=="PBAR") Flags.beam[1]=-1;
	else {
	  cout <<"Error in input file at line "<<iline<<endl;
	  cout <<"Unrecognised beam type for beam 2!"<<endl;
	  return 1;
	}
      }
      else if(label=="jetalgo:") {
	if(!(sline>>temp)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"jetalgo type invalid!"<<endl;
	  return 1;
	}
	if(temp=="antikt") Flags.jetalgo=fastjet::antikt_algorithm;
	else if(temp=="kt") Flags.jetalgo=fastjet::kt_algorithm;
	else {
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "Unrecognised beam type for beam 1!"<<endl;
	  return 1;
	}
      }
      else if(label=="Rparam:") {
	if(!(sline>>Flags.Rparam)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of Rparam invalid!"<<endl;
	  return 1;
	}
	if(Flags.Rparam<0) {
	  cout<<"Rparam<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="LHoutput:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "LHoutput flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.LHoutput=true; else Flags.LHoutput=false;
      }
      else if(label=="LHoutputfile:") {
	if(!(sline>>Flags.LHoutputfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "Eventfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="SubtractFirstOnly:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "LHoutput flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") SubtractFirstOnly=true; else SubtractFirstOnly=false;
      }
      else if(label=="rootfile:") {
	if(!(sline>>Flags.rootfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "rootfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="match:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.match=true; else Flags.match=false;
      }
      else if(label=="resum:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.resum=true; else Flags.resum=false;
      }
      else if(label=="nonFKLmatch2j:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.nonFKLmatch2j=true; else Flags.nonFKLmatch2j=false;
      }
      else if(label=="nonFKLmatch3j:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.nonFKLmatch3j=true; else Flags.nonFKLmatch3j=false;
      }
      else if(label=="nonFKLmatch4j:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "match flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.nonFKLmatch4j=true; else Flags.nonFKLmatch4j=false;
      }
      else if(label=="rootevents:") {
      	if(!(sline>>temp)){
      	  cout << "Error in input file at line "<<iline<<endl;
      	  cout << "rootevents flag invalid!"<<endl;
      	  return 1;
      	}
      	if(temp=="Y"||temp=="y") Flags.rootevents=true; else Flags.rootevents=false;
      }
      else if(label=="shower:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "shower flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") {Flags.shower=true; Flags.LHoutput=true;} else Flags.shower=false;
      }
      else if(label=="hadronise:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "hadronise flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") {Flags.shower=true; Flags.hadronise=true; Flags.LHoutput=true;} else Flags.hadronise=false;
      }
      else if(label=="unweight:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "unweight flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.unweight=true; else Flags.unweight=false;
      }
      else if(label=="mstwpdffile:") {
	if(!(sline>>Flags.mstwpdffile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "mstwpdffile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="LHAPDFsubset:") {
	if(!(sline>>Flags.LHAPDFsubset)){
	  cout<<"Error in input file at line "<<iline<<endl;
	  cout<<"Value of LHAPDFsubset invalid!"<<endl;
	  return 1;
	}
	if(Flags.LHAPDFsubset<0) {
	  cout<<"LHAPDFsubset<0 at line "<<iline<<" of input file!"<<endl;
	  return 1;
	}
      }
      else if(label=="LHAPDFName:") {
	if(!(sline>>Flags.LHAPDFName)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "LHAPDFName name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label=="useLHApdf:") {
	if(!(sline>>temp)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "useLHApdf flag invalid!"<<endl;
	  return 1;
	}
	if(temp=="Y"||temp=="y") Flags.useLHApdf=true; else Flags.useLHApdf=false;
      }
      else if(label=="raninitfile:") {
	if(!(sline>>Flags.raninitfile)){
	  cout << "Error in input file at line "<<iline<<endl;
	  cout << "raninitfile name invalid!"<<endl;
	  return 1;
	}
      }
      else if(label.at(0)==47){
      }
      else{
	cout<<"Unrecognised parameter at line "<<iline<<" in input file:"<<endl;
	cout<<line<<endl;
	return 1;
      }
    }
    
    input.close();
  }
  fastjet::Strategy strategy = fastjet::Best;
  fastjet::RecombinationScheme recomb_scheme = fastjet::E_scheme;
  Flags.jet_def=fastjet::JetDefinition(Flags.jetalgo, Flags.Rparam, recomb_scheme, strategy);

  if (Flags.shower) {
    int p1 = Flags.beam[0]*2212;
    int p2 = Flags.beam[1]*2212;
    double ecm = 2.*Flags.Ebeam;
    int hadr = 0;       // Set to 1 to get hadronization and decays
    
    // Initialize Ariadne and Pythia, giving the incoming particle types
    // and cm-energy.
    if (Flags.hadronise) 
      hadr=1;

    double hejptcut = Flags.extpartonptmin;
    arjini_(&p1, &p2, &ecm, &hadr, &hejptcut); 
    
    if (Flags.hadronise) {
      // *** Uncomment this if you want hadronization ***
      int isw = 5;
      int ival = 1;
      setmsta_(&isw, &ival);
    }

    // *** Uncomment this to avoid millions of photons from pi0-decay.
    int ip = 111;
    int ist = 0;
    setstable_(&ip, &ist);
    // ip = 25;
    // setstable_(&ip, &ist);

  }
  return 0;
}

void Fill2DHistograms(TH1D* h, TH1D* hN, TH1D *hwt,double x,double wt)
{
  double binwidth; // the binwidth
  binwidth=h->GetXaxis()->GetBinWidth(1);
  h->Fill(x,wt/binwidth);
  hwt->Fill(x,wt*wt/binwidth);
  hN->Fill(x);
}

void Fill2DHistograms(TH1D* h,double x,double wt)
{
  double binwidth; // the binwidth
  binwidth=h->GetXaxis()->GetBinWidth(1);
  h->Fill(x,wt/binwidth);
}

int main(int argn, char *argv[])
{
  double wt,wts; // The event weight
  // double ya,yb;
  CLHEP::HepLorentzVector pH,pi;
  int np; // The number of final state partons
  int snp; // The number of final state partons
  int njetsb,njetst; // The number of final state jets
  CLHEP::HepLorentzVector ja,jb,jthird,jfourth,jha,jhb;
  CLHEP::Ranlux64Engine rUnweight;
  
  // Float_t rpp[24][4]; // four vectors for parton momenta, for use with ROOT
  // Float_t rpH[4]; // the four vector for the Higgs boson, for use with ROOT

  if ((init(argn,argv))==1) 
    return(-1);
  // Only simple analysis and self-consistency check:
  // Example histogram file:
  TFile hfile(Flags.rootfile.c_str(),"RECREATE","Output from JetGen");
  TROOT simple("JetGen","Output from JetGen");
  TH1D *hmstatus,*hwt,*hwtwt;
  TH1D *htotal,*htotalN,*htotalwt2;
  htotal = new TH1D("total","total",1,-0.5,.5);
  htotal->Sumw2();
  htotalN = new TH1D("totalN","totalN",1,-0.5,.5);
  htotalwt2 = new TH1D("totalwt2","totalwt2",1,-0.5,.5);

  hmstatus = new TH1D("mstatus","mstatus",25,-.5,24.5);
  hwt = new TH1D("wt","wt",800,-30.,50.);
  hwtwt = new TH1D("wtwt","wtwt",800,-30.,50.);

  // Make tree of events:
  Float_t rpp[22][4]; // four vectors for parton momenta, for use with ROOT
  Float_t srpp[22][4]; // four vectors for parton momenta, for use with ROOT

  TTree * Events;
  Events = new TTree("Events","Events");
  Events->Branch("weight",&wts,"weights/D");
  Events->Branch("np",&np,"np/I");
  Events->Branch("snp",&snp,"snp/I");
  Events->Branch("Pfourv",&rpp,"Pfourv[np][4]/F");
  Events->Branch("SPfourv",&srpp,"SPfourv[snp][4]/F");
  std::cout<<"# ";
  CMultijet event(Flags); // use just standard settings
  generator=&event;
  Events->Branch("pdfta",&event.pdfta,"pdfta/D");
  Events->Branch("pdftb",&event.pdftb,"pdftb/D");

  for (unsigned int i=0;i<Flags.NEvents;i++) {
    bool savehej(false),saveshower(false);
    wt=event.Event(); // in micro-barn
    wts=wt/double(Flags.NEvents); // in micro-barn
    hmstatus->Fill(event.getMstatus());
    if (event.getMstatus()==0) { // If event was good
      // zero the momentum info for the HEJ event root file:
      for (int ii=0;ii<22;ii++) { // reset the rest
	rpp[ii][0]=0.;
	rpp[ii][1]=0.;
	rpp[ii][2]=0.;
	rpp[ii][3]=0.;
      }
      // zero the momentum info for the showered root file:
      for (int ii=0;ii<22;ii++) { // reset the rest
	srpp[ii][0]=0.;
	srpp[ii][1]=0.;
	srpp[ii][2]=0.;
	srpp[ii][3]=0.;
      }

      // Number of partons in event:
      np=event.getnp();
      
      std::vector<fastjet::PseudoJet> avec;
      
      double maxppt(0.); // the maximum partonic transverse momentum
      for (int ii=0;ii<np;ii++) {
	pi=event.getpartvn(ii);
	avec.push_back(pi);
	if (pi.perp()>maxppt)
	  maxppt=pi.perp();
      }
	
      /** fastjet find clusters */
      fastjet::ClusterSequence ev(avec, Flags.jet_def);
      
      /** Retrieve the final state jets from fastjet Sorted by Rapidity*/
      std::vector<fastjet::PseudoJet> softjets = sorted_by_rapidity(ev.inclusive_jets(Flags.genjetptmin));
      
      njetsb=int(softjets.size());

      if (njetsb>=2) {
	// Now perform the analysis on the pure HEJ event
	// Count how many of the fastjet jets have Et>energy_cut GeV 
	// and calculate largest distance between jets of more than ptcut2
	// now check if we have the required hard jets
	std::vector<fastjet::PseudoJet> alljets = sorted_by_pt(ev.inclusive_jets(Flags.jetptmin));

	// Make a list of those jets with rapidity less than Flags.jetrapmax
	std::vector<fastjet::PseudoJet> jets;
	{
	  std::vector<fastjet::PseudoJet>::const_iterator itr = alljets.begin();
	  for( ; itr != alljets.end() ; ++itr) {
	    if (fabs((*itr).rapidity())<jetrapmax)
	      jets.push_back(*itr);
	  }
	}
	jets=sorted_by_rapidity(jets);

	njetst=jets.size();
	  
	double ptbar(0.);
	if (njetst>=2) 
	  ptbar=(jets.at(0).perp()+jets.at(1).perp())/2.;
	
	if ((njetst>=2) && (ptbar > ptbarmin) ) {// A maximum selection, including both A and B
	  // Perform the analysis
	  event.xsec[event.getprocesstype()][0]+=wts;
	  event.xsec[event.getprocesstype()][1]+=wts*wts;
	  if (wt>event.maxweight[event.getprocesstype()])
	    event.maxweight[event.getprocesstype()]=wt;
	    
	  
	  if (Flags.rootevents) {
	    if ((Flags.unweight&&(rUnweight.flat()<wt/event.maxweight[event.getprocesstype()]))||!Flags.unweight) {
	      if (np>22) std::cout << "Ouch!\n";
	      // save jet information instead of partonic information
	      int ii=0;
	      std::vector<fastjet::PseudoJet>::const_iterator itr = jets.begin();
	      for( ; itr != jets.end() ; ++itr) {
		rpp[ii][0]=(*itr).e();
		rpp[ii][1]=(*itr).px();
		rpp[ii][2]=(*itr).py();
		rpp[ii][3]=(*itr).pz();
		++ii;
	      }
	      // for (int ii=0;ii<np;ii++) {
	      // 	pi=event.getpartvn(ii);
	      // 	rpp[ii][0]=pi.e();
	      // 	rpp[ii][1]=pi.x();
	      // 	rpp[ii][2]=pi.y();
	      // 	rpp[ii][3]=pi.z();
	      // }
	      // for (int ii=np;ii<22;ii++) { // reset the rest
	      // 	rpp[ii][0]=0.;
	      // 	rpp[ii][1]=0.;
	      // 	rpp[ii][2]=0.;
	      // 	rpp[ii][3]=0.;
	      // }
	      // Events->Fill();
	      np=jets.size();
	      savehej=true;
	    }
	  }
	    
	  // std::cout <<pe->perp() << " "<<pe->rapidity()<<" "<<pnu->perp()<<std::endl;
	  // htotal->Fill(0.,wts);
	  Fill2DHistograms(htotal,htotalN,htotalwt2,0.,wts);
	    
	  hwt->Fill(log(wt));
	  hwtwt->Fill(log(wt),wt);
	}
      } // end njetsb >=2
      if (Flags.shower) { //shower the event and perform the analysis on the aftermath
	//	  std::cerr <<"shower, ";
	// std::cout << "Shower and s(h)ave!\n";
	// First we declare the original state in Les Houches
	// format. Should begin with the two incoming partons, followed by
	// the leading jet corresponding to the first incoming one, then
	// all emitted gluons (in rapidity order?), and ending with the
	// leading jet corresponding to the second incoming partons.
	double pi[maxp][5]; // momenta
	int ifl[maxp];      // particle id
	int ico[maxp];      // colour-line index
	int iac[maxp];      // anti colour line index
	  
	// Then the final-state partons after showering. This will also
	// contain remnant particles. Optionally this can be the full
	// final-state after hadronization and decay.
	double po[maxp][5]; // momenta
	int ofl[maxp];      // particle id
	int oco[maxp];      // colour-line index
	int oac[maxp];      // anti colour line index
	  
	std::vector<particle> Particles=event.GetRedLHParticles();
	for (unsigned int ii=0;ii<Particles.size();ii++) {
	  particle temp=Particles.at(ii);
	  ifl[ii]=temp.PID;
	  ico[ii]=temp.COL1;
	  iac[ii]=temp.COL2;
	  pi[ii][0]=temp.p.px();
	  pi[ii][1]=temp.p.py();
	  pi[ii][2]=temp.p.pz();
	  pi[ii][3]=temp.p.e();
	  pi[ii][4]=temp.p.m();
	    
	}
	int npi = Particles.size();
	int npo = 0;
	  
	// Call and analyse the ariadne shower 
	arjshw_(&maxppt, &npi, pi[0], ifl, ico, iac, &maxp,
		&npo, po[0], ofl, oco, oac);
	if ( npo == 0 ) {
	  std::cerr <<"*** shower failed***" << std::endl;
	  continue;
	}
	//	  else std::cerr <<"shower finished, ";
	
	// Now jet cluster the event from shower
	std::vector<fastjet::PseudoJet> avec;
	  
	for (int iii=0;iii<npo;iii++) {
	  CLHEP::HepLorentzVector p;
	  p.set(po[iii][0],po[iii][1],po[iii][2],po[iii][3]);
	  avec.push_back(p); // add the particles that should
	  // enter jet clustering
	}
	  
	fastjet::ClusterSequence evA(avec, Flags.jet_def);
	// std::cout<< "Number of Hard Shower jets: "<<jetsA.size()<<endl;
	// loop over the constituents of each jet and calculate
	// the energy profile of jet
	// Now re-analyse everything on the aftermath of the shower
	{ // Now perform the analysis on the shower event

	  // Count how many of the fastjet jets have Et>energy_cut GeV 
	  // and calculate largest distance between jets of more than ptcut2
	  // now check if we have the required hard jets
	  // find the two hardest jets
	  std::vector<fastjet::PseudoJet> alljets = sorted_by_pt(evA.inclusive_jets(Flags.jetptmin));
	  
	  // Make a list of those jets with rapidity less than Flags.jetrapmax
	  std::vector<fastjet::PseudoJet> jets;
	  {
	    std::vector<fastjet::PseudoJet>::const_iterator itr = alljets.begin();
	    for( ; itr != alljets.end() ; ++itr) {
	      if (fabs((*itr).rapidity())<jetrapmax)
		jets.push_back(*itr);
	    }
	  }
	  jets=sorted_by_pt(jets);
	  njetst=int(jets.size());
	  if (njetst>=2) {
	    ja.set(jets.at(0)[0],jets.at(0)[1],jets.at(0)[2],jets.at(0)[3]);
	    jb.set(jets.at(1)[0],jets.at(1)[1],jets.at(1)[2],jets.at(1)[3]);
	    // ya=ja.rapidity();
	    // yb=jb.rapidity();
	    
	    double ptbar=(ja.perp()+jb.perp())/2.;
	      
	    if (ptbar > ptbarmin) {
	      if (Flags.rootevents) {
		if ((Flags.unweight&&(rUnweight.flat()<wt/event.maxweight[event.getprocesstype()]))||!Flags.unweight) {
		  if (np>22) { std::cout << "Ouch!\n"; continue;}
		    int ii=0;
		    std::vector<fastjet::PseudoJet>::const_iterator itr = jets.begin();
		    for( ; itr != jets.end() ; ++itr) {
		      srpp[ii][0]=(*itr).e();
		      srpp[ii][1]=(*itr).px();
		      srpp[ii][2]=(*itr).py();
		      srpp[ii][3]=(*itr).pz();
		      ++ii;
		    }
		  // for (int ii=0;ii<npo;ii++) {
		  //   srpp[ii][0]=po[ii][0];
		  //   srpp[ii][1]=po[ii][1];
		  //   srpp[ii][2]=po[ii][2];
		  //   srpp[ii][3]=po[ii][3];
		  // }
		  // for (int ii=np;ii<22;ii++) { // reset the rest
		  // 	rpp[ii][0]=0.;
		  // 	rpp[ii][1]=0.;
		  // 	rpp[ii][2]=0.;
		  // 	rpp[ii][3]=0.;
		  // }
		  // Events->Fill();
		    snp=jets.size();
		  saveshower=true;
		}
	      }
	    }
	  }
	}  //End shower analysis
      }  // end if shower
      if (savehej||saveshower)
	Events->Fill();
    } // end Mstatus
  } // end loop over events
  std::cout<<"Cross section passing cuts (mub): "<<htotal->GetBinContent(1)<<" +/- "<<htotal->GetBinError(1)<<std::endl;
  for (int count=1;count<=4;count++) 
    std::cout<<"Cross section passing cuts (mub): "<<event.xsec[count][0]<<" +/- "<<sqrt(event.xsec[count][1])<<std::endl;

  for (int count=1;count<=4;count++) 
    std::cout << event.maxweight[count] << " ";
  std::cout << endl;

  hfile.Write();
  hfile.Close();
  return 0;
}
